
--[[

SM: State Manager
dsm/DSM: Device SM (state management for one light bank)

--]]

local SM={} -- StateManager

-- We must remove any jobs if the internal tables are updated
local function removeJobs(self)
   self.jobsT={} -- Remove all jobs, if any
end

-- Sort Lua table (list of light banks) based on the light bank's
-- priority set in admin/.domain.lsp
local function dsmListComp(dsm1,dsm2)
   return dsm1.priority < dsm2.priority
end

-- The various jobs the SM can manage

local function runWaveOff(self)
   for _,dsm in ipairs(self.list) do
      for ix,curState in ipairs(dsm.curStateT) do
         if curState == true and dsm.enabledT[ix] then
            self.setLightCb(dsm.peerT,ix,false)
            coroutine.yield()
         end
      end
   end
end

local function runWaveOn(self)
   for _,dsm in ipairs(self.list) do
      for ix,curState in ipairs(dsm.curStateT) do
         if curState == false and dsm.enabledT[ix] then
            self.setLightCb(dsm.peerT,ix,true)
            coroutine.yield()
         end
      end
   end

end

-- runWaveOn > runWaveOff or runWaveOff > runWaveOn
local function runWave(self)
   local off,on=0,0
   for _,dsm in ipairs(self.list) do
      for ix,curState in ipairs(dsm.curStateT) do
         if dsm.enabledT[ix] then
            if curState then
               on = on + 1
            else
               off = off +1
            end
         end
      end
   end
   if off > 0 or on > 0 then
      if on > off then
         runWaveOff(self)
         runWaveOn(self)
      else
         runWaveOn(self)
         runWaveOff(self)
      end
   end
end

local function runPsychedelic(self)
   local count=0
   for _,dsm in ipairs(self.list) do
      for enabled in ipairs(dsm.enabledT) do
         if enabled then count = count+1 end
      end
   end
   if count > 2 then
      for i=1,count*2 do
         local dsm = self.list[ba.rnd(1,#self.list)]
         local ix = ba.rnd(1,#dsm.enabledT)
         if dsm.enabledT[ix] then
            self.setLightCb(dsm.peerT,ix,not dsm.curStateT[ix])
            coroutine.yield(true)
         end
      end
   end
end

-- Return a list of available job types. Must match "actionT" below.
local function getCommandList()
   return {
      {cmd="wave", name="Wave"},
      {cmd="waveon", name="Wave On"},
      {cmd="waveoff", name="Wave Off"},
      {cmd="psychedelic", name="Psychedelic"},
   }
end

-- Lua version of "switch" statement
local actionT={
   wave=runWave,
   waveoff=runWaveOff,
   waveon=runWaveOn,
   psychedelic=runPsychedelic,
}

-- Create a light control job.
-- action: job name must be one of the names in 'actionT'
function SM:run(action)
   action = actionT[action] -- Switch statement
   if action then
      local jobs=0
      for k,v in pairs(self.jobsT) do jobs = jobs + 1 end
      if jobs < 8 then -- Max 8 jobs at any time
         local co = coroutine.create(function() action(self) end)
         self.jobsT[co]=co -- Add to job list
      end
   end
end

-- Add a DSM (light bank device SM). Called by .preload when a new device connects
-- peerT: the table returned by smq.etid2peer(etid)
-- cfgT: The device configuration table
function SM:add(peerT, cfgT)
   -- device SM (dsm)
   local dsm={peerT=peerT, curStateT={}, enabledT={}, transitionT={}}
   assert(not self.etidT[peerT.tid])
   removeJobs(self)
   self.etidT[peerT.tid] = dsm
   local devT=cfgT.devices[peerT.dkey]
   for ix,lightT in ipairs(devT.lightsT) do
      dsm.enabledT[ix]=lightT.enabled
      dsm.curStateT[ix]=false -- Default: off
      dsm.transitionT[ix]=os.time()
   end
   dsm.priority=devT.priority
   table.insert(self.list, dsm)
   table.sort(self.list, dsmListComp) 
end

-- Remove DSM. Called by .preload when a device disconnects
-- etid: device (light bank's) ephemeral topic ID.
function SM:remove(etid)
   assert(self.etidT[etid])
   removeJobs(self)
   local dsm=self.etidT[etid]
   self.etidT[etid]=nil
   for ix, dsmix in ipairs(self.list) do
      if dsm == dsmix then
         table.remove(self.list, ix)
         return
      end
   end
   assert(false, "not found")
end

-- Called if the administrator (via admin/.domain.lsp) updates the config file
-- peerT: the table returned by smq.etid2peer(etid)
-- cfgT: The device configuration table
function SM:updateCfg(peerT, cfgT)
   local dsm=self.etidT[peerT.tid]
   removeJobs(self)
   local devT=cfgT.devices[peerT.dkey]
   local enabledT={}
   for ix,lightT in ipairs(devT.lightsT) do
      enabledT[ix]=lightT.enabled
   end
   dsm.enabledT=enabledT
   dsm.priority=devT.priority
   table.sort(self.list, dsmListComp)
end

-- hour:min:sec, where sec is optional
local function parseTime(t)
   local h,m,s=t:match("(%d+):(%d+):?(%d*)")
   if h then
      t = 60*60*tonumber(h) + 60*tonumber(m)
      if s and #s > 0 then t = t+tonumber(s) end
      return t
   end
end

-- Called if home owner (via config/index.lsp) changes the time
function SM:setTime(cfgT)
   local t = {}
   for k,v in pairs(cfgT.time) do
      t[k] = parseTime(v)
   end
   self.timeT=t
   self.timeZoneOffset=cfgT.timeZoneOffset
end

-- The SMQ client application in .preload calls this function when a
-- device publishes light state change.
function SM:setLight(etid,lightIx,state)
   local dsm=self.etidT[etid]
   dsm.curStateT[lightIx]=state
   dsm.transitionT[lightIx]=os.time()
end

-- Returns the light bank's state(s). Each bank may have 1 to N lights.
function SM:getLightStates(etid)
   return self.etidT[etid].curStateT
end

-- Returns a 24 hour time value in seconds
local function getTime(self)
   local now = os.date("*t",os.time())
   local t = now.hour * 3600 + now.min * 60 + now.sec
   if self.timeZoneOffset then
      t = t+self.timeZoneOffset
      local oneDay=3600*24
      if t < 0 then return t+oneDay end
      if t > oneDay then return t-oneDay end
   end
   return t
end

--Calculate and return the on/off (true/false) state for the 'on' to 'off' time
local function calcState(self,onTime,offTime)
   local now=getTime(self)
   if onTime > offTime then
      return now >= onTime or now < offTime
   end
   return now >= onTime and now < offTime
end

-- Global light on/off state based on time set in /config/index.lsp
function SM:getLightOnState()
   return calcState(self,self.timeT.setOn,self.timeT.setOff)
end

-- Returns the enable/disable (true/false) state for public access to
-- the main index page.
function SM:getPubAccessState()
   return calcState(self,self.timeT.notBefore,self.timeT.notAfter)
end

-- Background timer function manages the light states and runs active jobs.
local function runJobs(self)
   local counter=0;
   local lightState = self:getLightOnState()
   while true do

      -- Check the lights default state every 8'th tick
      -- Set lights back to default state if time since last change > 60 secs
      if counter == 8 then
         local now = os.time()
         for _,dsm in ipairs(self.list) do
            for ix,curState in ipairs(dsm.curStateT) do
               if curState ~= lightState and
                  dsm.enabledT[ix] and
                  now - dsm.transitionT[ix] > 60 then
                  self.setLightCb(dsm.peerT,ix,lightState)
               end
            end
         end
      end

      -- Check the global light on/off state every 10'th tick and reset counter
      if counter >= 10 then
         counter = 0
         local nls = self:getLightOnState()
         if lightState ~= nls then -- Change state
            lightState = nls
            self:run(nls and "waveon" or "waveoff")
         end
      end
      counter = counter+1

      -- Run jobs, if any
      local count=0
      for co in pairs(self.jobsT) do
         if not coroutine.resume(co) then self.jobsT[co] = nil end
         count = count + 1
         if count > 3 then break end -- Run max three concurrent jobs
      end

      coroutine.yield(true) -- Wait for next timer tick
   end
end

-- Return API used by .preload
return {
   -- Create a Light Manager instance
   create=function(cfgT,setLightCb)
      local sm={
         etidT={},
         jobsT={},
         list={},
         setLightCb=setLightCb,
      }
      sm.runJobsTimer=ba.timer(function() runJobs(sm) end)
      sm.runJobsTimer:set(400)
      sm=setmetatable(sm, {__index=SM})
      sm:setTime(cfgT)
      return sm
   end,
   getCommandList=getCommandList,
}
